/****************************************************************************
 *									    *
 *									    *
 *   VMS Randomness-Gathering Code, David Mathog, Caltech		    *
 *   based (slightly) on Unix Randomness-Gathering Code                     *
 *   and strongly on the sshrsa_init() function from Dave Jones SSH server  *
 *   for OpenVMS.  (Ok, copied almost verbatim in sections!)   The copyright 
     for that distribution is:


        Copyright 1998,1999 The Ohio State University.
        The Ohio State University will not assert copyright with respect
        to reproduction,  distribution,  performance and/or modification
        of this program by  any person or  entity that  ensures that all
        copies made,  controlled or distributed by or for him or it bear
        appropriate acknowlegement of the developers of this program.

 *
 *									    *
 *   Copyright Peter Gutmann, Paul Kendall, and Chris Wedgwood 1996-1999.   *
 *   Heavily modified for GnuPG by Werner Koch				    *
 *									    *
 *									    *
 ****************************************************************************/

/* This module is part of the cryptlib continuously seeded pseudorandom
   number generator.  For usage conditions, see lib_rand.c

   [Here is the notice from lib_rand.c:]

   This module and the misc/rnd*.c modules represent the cryptlib
   continuously seeded pseudorandom number generator (CSPRNG) as described in
   my 1998 Usenix Security Symposium paper "The generation of random numbers
   for cryptographic purposes".

   The CSPRNG code is copyright Peter Gutmann (and various others) 1996,
   1997, 1998, 1999, all rights reserved.  Redistribution of the CSPRNG
   modules and use in source and binary forms, with or without modification,
   are permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice
      and this permission notice in its entirety.

   2. Redistributions in binary form must reproduce the copyright notice in
      the documentation and/or other materials provided with the distribution.

   3. A copy of any bugfixes or enhancements made must be provided to the
      author, <pgut001@cs.auckland.ac.nz> to allow them to be added to the
      baseline version of the code.

  ALTERNATIVELY, the code may be distributed under the terms of the GNU
  General Public License, version 2 or any later version published by the
  Free Software Foundation, in which case the provisions of the GNU GPL are
  required INSTEAD OF the above restrictions.

  Although not required under the terms of the GPL, it would still be nice if
  you could make any changes available to the author to allow a consistent
  code base to be maintained */



#include <unistd.h>
#include <config.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <descrip.h>                    /* VMS string descriptors */
#include <iodef.h>                      /* VMS QIO function codes */
#include <lckdef.h>                     /* VMS lock manager codes */
#include <jpidef.h>                     /* VMS JPI definitions */
#include <ssdef.h>                      /* VMS status definitions */
#include <sys/types.h>
#include "dynload.h"
#include "util.h"




#define GATHER_BUFSIZE		49152	/* Not much gets filled on VMS */

/* The structure containing information on random-data sources.  Each
 * record contains the source and a relative estimate of its usefulness
 * (weighting) which is used to scale the number of kB of output from the
 * source (total = data_bytes / usefulness).  Usually the weighting is in the
 * range 1-3 (or 0 for especially useless sources), resulting in a usefulness
 * rating of 1...3 for each kB of source output (or 0 for the useless
 * sources).
 *
 * If the source is constantly changing (certain types of network statistics
 * have this characteristic) but the amount of output is small, the weighting
 * is given as a negative value to indicate that the output should be treated
 * as if a minimum of 1K of output had been obtained.  If the source produces
 * a lot of output then the scale factor is fractional, resulting in a
 * usefulness rating of < 1 for each kB of source output.
 *
 * In order to provide enough randomness to satisfy the requirements for a
 * slow poll, we need to accumulate at least 20 points of usefulness (a
 * typical system should get about 30 points).
 *
 * Some potential options are missed out because of special considerations.
 * pstat -i and pstat -f can produce amazing amounts of output (the record
 * is 600K on an Oracle server) which floods the buffer and doesn't yield
 * anything useful (apart from perhaps increasing the entropy of the vmstat
 * output a bit), so we don't bother with this.  pstat in general produces
 * quite a bit of output, but it doesn't change much over time, so it gets
 * very low weightings.  netstat -s produces constantly-changing output but
 * also produces quite a bit of it, so it only gets a weighting of 2 rather
 * than 3.  The same holds for netstat -in, which gets 1 rather than 2.
 *
 * Some binaries are stored in different locations on different systems so
 * alternative paths are given for them.  The code sorts out which one to
 * run by itself, once it finds an exectable somewhere it moves on to the
 * next source.  The sources are arranged roughly in their order of
 * usefulness, occasionally sources which provide a tiny amount of
 * relatively useless data are placed ahead of ones which provide a large
 * amount of possibly useful data because another 100 bytes can't hurt, and
 * it means the buffer won't be swamped by one or two high-output sources.
 * All the high-output sources are clustered towards the end of the list
 * for this reason.  Some binaries are checked for in a certain order, for
 * example under Slowaris /usr/ucb/ps understands aux as an arg, but the
 * others don't.  Some systems have conditional defines enabling alternatives
 * to commands which don't understand the usual options but will provide
 * enough output (in the form of error messages) to look like they're the
 * real thing, causing alternative options to be skipped (we can't check the
 * return either because some commands return peculiar, non-zero status even
 * when they're working correctly).
 *
 * In order to maximise use of the buffer, the code performs a form of run-
 * length compression on its input where a repeated sequence of bytes is
 * replaced by the occurrence count mod 256.  Some commands output an awful
 * lot of whitespace, this measure greatly increases the amount of data we
 * can fit in the buffer.
 *
 * When we scale the weighting using the SC() macro, some preprocessors may
 * give a division by zero warning for the most obvious expression
 * 'weight ? 1024 / weight : 0' (and gcc 2.7.2.2 dies with a division by zero
 * trap), so we define a value SC_0 which evaluates to zero when fed to
 * '1024 / SC_0' */

/*
 * Utility function to seed random number generator from process state.
 * Taken pretty much verbatim from Dave Jones's SSH server.
 * returns the number of bytes placed into the buffer, up to
 * the value of "remaining" - the space left.
 */
int fill_ran_buffer(char *ranbuffer, int space)
{
    int  filled,toadd;
    long pid, iosb[2];
    int SYS$GETJPIW(), status, i, mbx_chan, SYS$GETTIM(), SYS$ENQW();
    static char lock_name_buf[64];
    static $DESCRIPTOR(lock_name,lock_name_buf);
    static long lksb[2];
    struct { short length, code; long *buffer; int *retlen; } item[10];
    struct { long pid, bufio, cputim, dio, pageflts, wssize, lgn[2]; } info;
#define SET_ITEM(A,B,C) item[A].length=sizeof(long); item[A].code=(B);\
        item[A].buffer = &info.C; item[A].retlen = (int *) 0;
    /*
     * Set up for JPI scan.
     */
    filled = 0;
    SET_ITEM(0,JPI$_PID,pid)
    SET_ITEM(1,JPI$_BUFIO,bufio)
    SET_ITEM(2,JPI$_DIRIO,dio)
    SET_ITEM(3,JPI$_CPUTIM,cputim)
    SET_ITEM(4,JPI$_PAGEFLTS,pageflts)
    SET_ITEM(5,JPI$_WSSIZE,wssize)
    SET_ITEM(6, JPI$_LOGINTIM, lgn[0])
    item[6].length = sizeof(info.lgn);
    item[7].length = item[7].code = 0;  /* terminate list */
    /*
     * Scan process list and load random number generator with results.
     */
    pid = -1;                   /* search context */
    for ( i = 0; ; i++ ) {
        status = SYS$GETJPIW ( 8, &pid,  0, item, iosb, 0, 0 );
        if ( status == SS$_NOMOREPROC ) break;
        if ( status == SS$_NOPRIV ) continue;
        if ( status == SS$_SUSPENDED ) continue;
        if ( status == SS$_NORMAL ) {
            toadd = sizeof(info)  - 1;
            if(toadd  > space)toadd=space;
            memcpy(ranbuffer,&info,toadd);
            filled += toadd;
            if(filled >= space)return filled;
        } else {
            printf("unexpected error in JPI call: %d\n", status );
            break;
        }
    }
    SYS$GETTIM ( iosb );
    toadd = sizeof(iosb);
    if(toadd - filled   > space)toadd=space - filled;
    memcpy(&ranbuffer[filled],&info,toadd);
    filled += toadd;
    return filled;
}


/****************
 * Using a level of 0 should never block and better add nothing
 * to the pool.  So this is just a dummy for this gatherer.
 */
static int
gather_random( void (*add)(const void*, size_t, int), int requester,
	       size_t length, int level )
{
char ranbuffer[GATHER_BUFSIZE];
    int n;

    if( !level )return 0;

    n = fill_ran_buffer(ranbuffer, GATHER_BUFSIZE);
    if( n > length )n = length;
    (*add)( ranbuffer, n, requester );
    return 0;
}



#ifndef IS_MODULE
static
#endif
const char * const gnupgext_version = "RNDVMS ($Revision: 1.14.2.3 $)";


static struct {
    int class;
    int version;
    void *func;
} func_table[] = {
    { 40, 1, gather_random },
};

/****************
 * Enumerate the names of the functions together with informations about
 * this function. Set sequence to an integer with a initial value of 0 and
 * do not change it.
 * If what is 0 all kind of functions are returned.
 * Return values: class := class of function:
 *			   10 = message digest algorithm info function
 *			   11 = integer with available md algorithms
 *			   20 = cipher algorithm info function
 *			   21 = integer with available cipher algorithms
 *			   30 = public key algorithm info function
 *			   31 = integer with available pubkey algorithms
 *			   40 = get read_random_source() function
 *			   41 = get fast_random_poll function
 *		  version = interface version of the function/pointer
 *			    (currently this is 1 for all functions)
 */

#ifndef IS_MODULE
static
#endif
void *
gnupgext_enum_func( int what, int *sequence, int *class, int *vers )
{
    void *ret;
    int i = *sequence;

    do {
	if ( i >= DIM(func_table) || i < 0 ) {
	    return NULL;
	}
	*class = func_table[i].class;
	*vers  = func_table[i].version;
	ret = func_table[i].func;
	i++;
    } while ( what && what != *class );

    *sequence = i;
    return ret;
}

#ifndef IS_MODULE
void
rndvms_constructor(void)
{
    register_internal_cipher_extension( gnupgext_version,
					gnupgext_enum_func );
}
#endif


